@@ -52,6 +52,7 @@ struct _RsvgPaintServer {
 
 struct _RsvgPaintServerSolid {
 	RsvgPaintServer super;
+	gboolean currentcolour;
 	guint32 rgb;
 };
 
@@ -84,6 +85,8 @@ rsvg_paint_server_solid_render (RsvgPaintServer *self, ArtRender *ar,
 {
 	RsvgPaintServerSolid *z = (RsvgPaintServerSolid *)self;
 	guint32 rgb = z->rgb;
+	if (z->currentcolour)
+		rgb = rsvg_state_current(ctx->ctx)->current_color;
 	ArtPixMaxDepth color[3];
 	
 	color[0] = ART_PIX_MAX_FROM_8 (rgb >> 16);
@@ -103,6 +106,21 @@ rsvg_paint_server_solid (guint32 rgb)
 	result->super.render = rsvg_paint_server_solid_render;
 	
 	result->rgb = rgb;
+	result->currentcolour = FALSE;
+	
+	return &result->super;
+}
+
+static RsvgPaintServer *
+rsvg_paint_server_solid_current_colour ()
+{
+	RsvgPaintServerSolid *result = g_new (RsvgPaintServerSolid, 1);
+	
+	result->super.refcnt = 1;
+	result->super.free = rsvg_paint_server_solid_free;
+	result->super.render = rsvg_paint_server_solid_render;
+	
+	result->currentcolour = TRUE;
 	
 	return &result->super;
 }
@@ -166,9 +184,9 @@ rsvg_paint_server_lin_grad_render (RsvgPaintServer *self, ArtRender *ar,
 	double affine[6];
 	guint32 current_color;
 	int i;
-	float xchange, ychange, pointlen,unitlen;
-	float nx2, ny2;
-	float x0, y0;
+	double xchange, ychange, pointlen,unitlen;
+	double nx2, ny2;
+	double x0, y0;
 
 	agl = z->agl;
 	if (agl == NULL)
@@ -195,7 +213,7 @@ rsvg_paint_server_lin_grad_render (RsvgPaintServer *self, ArtRender *ar,
 		affine[3] = ctx->y1 - ctx->y0;
 		affine[4] = ctx->x0;
 		affine[5] = ctx->y0;
-
+		art_affine_multiply(affine, affine, ctx->affine);
 	} else {
 		for (i = 0; i < 6; i++)
 			affine[i] = ctx->affine[i];
@@ -227,8 +245,12 @@ rsvg_paint_server_lin_grad_render (RsvgPaintServer *self, ArtRender *ar,
 	ychange = y2 - y1;
 	unitlen = sqrt(xchange*xchange + ychange*ychange);
 
-	x2 = x1 + ychange / unitlen * pointlen;
-	y2 = y1 - xchange / unitlen * pointlen;
+	if (unitlen == 0 || pointlen == 0) {
+		x2 = y2 = 0;
+	} else {
+		x2 = x1 + ychange / unitlen * pointlen;
+		y2 = y1 - xchange / unitlen * pointlen;
+	}
 
 	/* solve a, b, c so ax1 + by1 + c = 0 and ax2 + by2 + c = 1, maximum
 	   gradient is in x1,y1 to x2,y2 dir */
@@ -237,7 +259,7 @@ rsvg_paint_server_lin_grad_render (RsvgPaintServer *self, ArtRender *ar,
 
 	/* workaround for an evil devide by 0 bug - not sure if this is sufficient */
 	if (fabs(dx) + fabs(dy) <= 0.0000001)
-		scale = 0.;
+		scale = 100000000.;
 	else
 		scale = 1.0 / (dx * dx + dy * dy);
 	agl->a = dx * scale;
@@ -292,7 +314,7 @@ rsvg_paint_server_rad_grad_render (RsvgPaintServer *self, ArtRender *ar,
 		affine[3] = ctx->y1 - ctx->y0;
 		affine[4] = ctx->x0;
 		affine[5] = ctx->y0;
-
+		art_affine_multiply(affine, affine, ctx->affine);
 	} else {
 		for (i = 0; i < 6; i++)
 			affine[i] = ctx->affine[i];
@@ -349,10 +371,12 @@ rsvg_paint_server_rad_grad (RsvgRadialGradient *gradient)
 typedef struct {
 	ArtImageSource super;
 	gchar * pixels;
-	gint x, y, width, height;
+	gdouble x, y, width, height, xoffset, yoffset;
 	gint realwidth, realheight; 
 	gint rowstride;
 	art_boolean init;
+	gdouble affine[6];
+	gdouble invaffine[6];
 } RsvgImageSourcePattern;
 
 static void
@@ -364,43 +388,51 @@ render_image_pattern_done (ArtRenderCallback *self, ArtRender *render)
 	g_free(self);
 }
 
-#include <libart_lgpl/art_rgb.h>
+#include <libart_lgpl/art_rgba.h>
 #include <libart_lgpl/art_render.h>
 
-static int goodmod(int one, int two)
-{
-	while (one < 0)
-		one += two;
-	return one % two;
-}
+/*the commented out regions in the next bit allow overflow*/
 
 static void
 render_image_pattern_render(ArtRenderCallback *self, ArtRender *render,
 				  art_u8 *dest, int y)
 {
 	RsvgImageSourcePattern *z = (RsvgImageSourcePattern *)self;
-	int i;	
-
+	int i;
 	int x0 = render->x0;
 	int x1 = render->x1;
 
 	int sx, sy;
+	double px, py, gx, gy, gnx, gny, tx, ty;
 
+	tx = -z->x * z->affine[0] + -z->y * z->affine[2] + z->affine[4];
+	ty = -z->x * z->affine[1] + -z->y * z->affine[3] + z->affine[5];
 
-	for (i = 0; i < x1 - x0; i += 1)
+	for (i = 0; i < x1 - x0; i++)
 		{
-			sx = goodmod((i + x0 + z->x),z->width);
-			sy = goodmod((y + render->y0 + z->y),z->height);
-			if (sx < 0 || sx >= z->realwidth || sy < 0 || sy >= z->realheight)
-				{
-					render->image_buf[i * 4 + 3] = 0;
-					continue;
-				}
-			//printf("%i, %i -> %i, %i\n", i, y, sx, sy);
-			render->image_buf[i * 4] = z->pixels[sx * 4 + z->rowstride * sy];
+			px = i;
+			py = y;
+			
+			gx = px * z->invaffine[0] + py * z->invaffine[2] + z->invaffine[4] - z->x;
+			gy = px * z->invaffine[1] + py * z->invaffine[3] + z->invaffine[5] - z->y;
+			
+			gnx = floor (gx / z->width);
+			gny = floor (gy / z->height);
+	
+			sx = px - gnx * z->width * z->affine[0] - gny * z->height * z->affine[2] - z->affine[4]
+				+ z->xoffset + tx;
+			sy = py - gnx * z->width * z->affine[1] - gny * z->height * z->affine[3] - z->affine[5]
+				+ z->yoffset + ty;
+			
+			if (sx < 0 || sx >= z->realwidth || sy < 0 || sy >= z->realheight){
+				render->image_buf[i * 4 + 3] = 0;
+				continue;
+			}
+			render->image_buf[i * 4 + 0] = z->pixels[sx * 4 + z->rowstride * sy];
 			render->image_buf[i * 4 + 1] = z->pixels[sx * 4 + z->rowstride * sy + 1];
 			render->image_buf[i * 4 + 2] = z->pixels[sx * 4 + z->rowstride * sy + 2];
 			render->image_buf[i * 4 + 3] = z->pixels[sx * 4 + z->rowstride * sy + 3];
+			
 		}
 }
 
@@ -416,9 +448,9 @@ render_image_pattern_negotiate (ArtImageSource *self, ArtRender *render,
 }
 
 static void
-render_image_pattern (ArtRender *render, gchar * pixels, gint x, gint y, 
-					  gint width, gint height, gint realwidth, gint realheight, gint rowstride,
-					  double * affine)
+render_image_pattern (ArtRender *render, guchar * pixels, gdouble x, gdouble y, 
+					  gdouble width, gdouble height, gint realwidth, gint realheight, gint rowstride,
+					  gdouble xoffset, gdouble yoffset, double * affine)
 {	
 	RsvgImageSourcePattern *image_source;
 	int i;
@@ -437,10 +469,17 @@ render_image_pattern (ArtRender *render, gchar * pixels, gint x, gint y,
 	image_source->realheight = realheight;
 	image_source->x = x;
 	image_source->y = y;
+	image_source->xoffset = xoffset;
+	image_source->yoffset = yoffset;
 		
 	for (i = 0; i < rowstride * realheight; i++)
 		image_source->pixels[i] = pixels[i];  
 	
+	for (i = 0; i < 6; i++)
+		image_source->affine[i] = affine[i];  
+
+	art_affine_invert(image_source->invaffine, affine);
+
 	image_source->init = ART_FALSE;
 	
 	art_render_add_image_source (render, &image_source->super);
@@ -461,12 +500,16 @@ rsvg_paint_server_pattern_render (RsvgPaintServer *self, ArtRender *ar,
 	RsvgPaintServerPattern *z = (RsvgPaintServerPattern *)self;
 	RsvgPattern *pattern = z->pattern;
 	RsvgDefsDrawable *drawable = (RsvgDefsDrawable *)pattern->g;
-	RsvgHandle *hctx = ctx->ctx;
+	DrawingCtx *hctx = ctx->ctx;
 	double affine[6];
 	double caffine[6];
-	int i;
+	int i, j;
+	gdouble minx, miny, maxx, maxy, xcoord, ycoord, xoffset, yoffset;
 	GdkPixbuf *save, *render;
 
+	if (ctx->ctx == NULL)
+		return;
+
 	if (pattern->obj_bbox) {
 		affine[0] = ctx->x1 - ctx->x0;
 		affine[1] = 0.;		
@@ -474,47 +517,88 @@ rsvg_paint_server_pattern_render (RsvgPaintServer *self, ArtRender *ar,
 		affine[3] = ctx->y1 - ctx->y0;
 		affine[4] = ctx->x0;
 		affine[5] = ctx->y0;
-
+		art_affine_multiply(affine, affine, ctx->affine);
 	} else {
 		for (i = 0; i < 6; i++)
 			affine[i] = ctx->affine[i];
 	}
 
-	if (pattern->obj_cbbox) {
+	if (pattern->vbox) {
+		double w, h, x, y;
+		w = pattern->width;
+		h = pattern->height;
+		x = 0;
+		y = 0;
+
+		rsvg_preserve_aspect_ratio(pattern->preserve_aspect_ratio,
+								   pattern->vbw, pattern->vbh, 
+								   &w, &h, &x, &y);
+
+		x -= pattern->vbx * w / pattern->vbw;
+		y -= pattern->vby * h / pattern->vbh;
+
+		caffine[0] = w / pattern->vbw;
+		caffine[1] = 0.;		
+		caffine[2] = 0.;
+		caffine[3] = h / pattern->vbh;
+		caffine[4] = x;
+		caffine[5] = y;
+		art_affine_multiply(caffine, caffine, affine);
+	}
+	else if (pattern->obj_cbbox) {
 		caffine[0] = ctx->x1 - ctx->x0;
 		caffine[1] = 0.;		
 		caffine[2] = 0.;
 		caffine[3] = ctx->y1 - ctx->y0;
 		caffine[4] = ctx->x0;
 		caffine[5] = ctx->y0;
-
+		art_affine_multiply(caffine, caffine, ctx->affine);
 	} else {
 		for (i = 0; i < 6; i++)
 			caffine[i] = ctx->affine[i];
 	}
 
-	art_affine_multiply(affine, pattern->affine, affine);
-	art_affine_multiply(caffine, pattern->affine, caffine);
+	art_affine_multiply(affine, affine, pattern->affine);
+	art_affine_multiply(caffine, caffine, pattern->affine);
 
-	render = gdk_pixbuf_new (GDK_COLORSPACE_RGB, 1, 8, 
-							 gdk_pixbuf_get_width(hctx->pixbuf), 
-							 gdk_pixbuf_get_height(hctx->pixbuf));
+	/*check if everything is going to be within the boundaries of the rendering surface*/
+	maxx = maxy = minx = miny = xoffset = yoffset = 0;
 
-	gdk_pixbuf_fill(render, 0x00000000);	
+	for (i = 0; i < 2; i++)
+		for (j = 0; j < 2; j++)
+			{
+				xcoord = affine[0] * pattern->width * i + affine[2] * pattern->height * j + affine[4];
+				ycoord = affine[1] * pattern->width * i + affine[3] * pattern->height * j + affine[5];
+				if (xcoord < minx)
+					minx = xcoord;
+				if (xcoord > maxx)
+					maxx = xcoord;
+				if (ycoord < miny)
+					miny = ycoord;
+				if (ycoord > maxy)
+					maxy = ycoord;
+			}
+
+	if (minx < 0)
+		xoffset = -minx;
+	if (miny < 0)
+		yoffset = -miny;
+	if (maxx > gdk_pixbuf_get_width(hctx->pixbuf))
+		xoffset = -minx;
+	if (maxy > gdk_pixbuf_get_height(hctx->pixbuf))
+		yoffset = -maxy;	
+
+	render = _rsvg_pixbuf_new_cleared(GDK_COLORSPACE_RGB, 1, 8, 
+									  gdk_pixbuf_get_width(hctx->pixbuf), 
+									  gdk_pixbuf_get_height(hctx->pixbuf));
 	save = hctx->pixbuf;
 
 	hctx->pixbuf = render;
 
-	/* push the state stack */
-	if (hctx->n_state == hctx->n_state_max)
-		hctx->state = g_renew (RsvgState, hctx->state, 
-							   hctx->n_state_max <<= 1);
-	if (hctx->n_state)
-		rsvg_state_inherit (&hctx->state[hctx->n_state],
-							&hctx->state[hctx->n_state - 1]);
-	else
-				rsvg_state_init (hctx->state);
-	hctx->n_state++;
+	rsvg_state_push(ctx->ctx);
+	
+	caffine[4] += xoffset;
+	caffine[5] += yoffset;
 
 	for (i = 0; i < 6; i++)
 		{
@@ -528,20 +612,19 @@ rsvg_paint_server_pattern_render (RsvgPaintServer *self, ArtRender *ar,
 	else
 		rsvg_defs_drawable_draw ((RsvgDefsDrawable *)pattern->gfallback, hctx, 2);		
 
-	/* pop the state stack */
-	hctx->n_state--;
-	rsvg_state_finalize (&hctx->state[hctx->n_state]);
+	rsvg_state_pop(ctx->ctx);
 
   	hctx->pixbuf = save;
 
 	render_image_pattern (ar, gdk_pixbuf_get_pixels (render),
-						  pattern->x * affine[0] + affine[4], 
-						  pattern->y * affine[3] + affine[5], 
-						  pattern->width * affine[0], 
-						  pattern->height * affine[3], 
+						  pattern->x, pattern->y, 
+						  pattern->width, pattern->height, 
 						  gdk_pixbuf_get_width (render),
 						  gdk_pixbuf_get_height (render),
-						  gdk_pixbuf_get_rowstride (render), affine);
+						  gdk_pixbuf_get_rowstride (render), 
+						  xoffset, yoffset, affine);
+
+	g_object_unref(G_OBJECT(render));
 }
 
 static RsvgPaintServer *
@@ -569,11 +652,12 @@ rsvg_paint_server_pattern (RsvgPattern *pattern)
  * Return value: The newly created paint server, or NULL on error.
  **/
 RsvgPaintServer *
-rsvg_paint_server_parse (RsvgPaintServer * current, const RsvgDefs *defs, const char *str,
+rsvg_paint_server_parse (gboolean * inherit, const RsvgDefs *defs, const char *str,
 						 guint32 current_color)
 {
 	guint32 rgb;
-	
+	if (inherit != NULL)
+		*inherit = 1;
 	if (!strcmp (str, "none"))
 		return NULL;
 
@@ -585,9 +669,6 @@ rsvg_paint_server_parse (RsvgPaintServer * current, const RsvgDefs *defs, const
 			RsvgDefVal *val;
 			
 			while (g_ascii_isspace (*p)) p++;
-			if (*p != '#')
-				return NULL;
-			p++;
 			for (ix = 0; p[ix]; ix++)
 				if (p[ix] == ')') break;
 			if (p[ix] != ')')
@@ -609,20 +690,23 @@ rsvg_paint_server_parse (RsvgPaintServer * current, const RsvgDefs *defs, const
 					return NULL;
 				}
 		}
-	else if (current && !strcmp (str, "inherit"))
+	else if (!strcmp (str, "inherit"))
 		{
-			rsvg_paint_server_ref (current);
-			return current;
+			if (inherit != NULL)
+				*inherit = 0;
+			return rsvg_paint_server_solid (0);
+		}
+	else if (!strcmp (str, "currentColor"))
+		{	
+			RsvgPaintServer * ps;			
+			ps = rsvg_paint_server_solid_current_colour ();
+			return ps;
 		}
 	else
-	  {
-		  if (!strcmp (str, "currentColor"))
-			  rgb = current_color;
-		  else
-			  rgb = rsvg_css_parse_color (str, 0);
-
-		  return rsvg_paint_server_solid (rgb);
-	  }
+		{
+			rgb = rsvg_css_parse_color (str, inherit);
+			return rsvg_paint_server_solid (rgb);
+		}
 }
 
 /**
@@ -765,6 +849,7 @@ rsvg_clone_pattern (const RsvgPattern *pattern)
 	
 	clone->obj_bbox = pattern->obj_bbox;
 	clone->obj_cbbox = pattern->obj_cbbox;
+	clone->vbox = pattern->vbox;
 	for (i = 0; i < 6; i++)
 		clone->affine[i] = pattern->affine[i];
 
@@ -774,5 +859,14 @@ rsvg_clone_pattern (const RsvgPattern *pattern)
 	else
 		clone->gfallback = pattern->gfallback;		
 
+	clone->x = pattern->x;
+	clone->y = pattern->y;
+	clone->width = pattern->width;
+	clone->height = pattern->height;
+	clone->vbx = pattern->vbx;
+	clone->vby = pattern->vby;
+	clone->vbw = pattern->vbw;	
+	clone->vbh = pattern->vbh;
+
 	return clone;
 }
